import Foundation
import XCTest

// MARK: - Tests

class UnboxTests: XCTestCase {
    func testWithOnlyValidRequiredValues() {
        let dictionary = UnboxTestDictionaryWithAllRequiredKeysWithValidValues(false)
        let unboxed: UnboxTestMock? = Unbox(dictionary)
        XCTAssertNotNil(unboxed, "Failed to unbox valid dictionary")
        unboxed?.verifyAgainstDictionary(dictionary)
    }
    
    func testWithMissingRequiredValues() {
        let validDictionary = UnboxTestDictionaryWithAllRequiredKeysWithValidValues(false)
        
        for key in validDictionary.keys {
            var invalidDictionary = validDictionary
            invalidDictionary.removeValueForKey(key)
            
            let unboxed: UnboxTestMock? = Unbox(invalidDictionary)
            XCTAssertNil(unboxed, "Unbox did not return nil for an invalid dictionary")
        }
    }
    
    func testWithInvalidRequiredValues() {
        let validDictionary = UnboxTestDictionaryWithAllRequiredKeysWithValidValues(false)
        
        for key in validDictionary.keys {
            var invalidDictionary = validDictionary
            invalidDictionary[key] = NSObject()
            
            let unboxed: UnboxTestMock? = Unbox(invalidDictionary)
            XCTAssertNil(unboxed, "Unbox did not return nil for an invalid dictionary")
        }
    }
    
    func testWithInvalidRequiredURL() {
        var invalidDictionary = UnboxTestDictionaryWithAllRequiredKeysWithValidValues(false)
        invalidDictionary[UnboxTestMock.requiredURLKey] = "Clearly not a URL!"
        
        let unboxed: UnboxTestMock? = Unbox(invalidDictionary)
        XCTAssertNil(unboxed, "Unbox did not return nil for a dictionary with an invalid required URL value")
    }
    
    func testWithInvalidRequiredUnboxable() {
        var invalidDictionary = UnboxTestDictionaryWithAllRequiredKeysWithValidValues(false)
        invalidDictionary[UnboxTestMock.requiredUnboxableKey] = "Totally not unboxable"
        
        let unboxedFromString: UnboxTestMock? = Unbox(invalidDictionary)
        XCTAssertNil(unboxedFromString, "Unbox did not return nil for a string")
        
        invalidDictionary[UnboxTestMock.requiredUnboxableKey] = ["cannotBe" : "unboxed"]
        
        let unboxedFromInvalidDictionary: UnboxTestMock? = Unbox(invalidDictionary)
        XCTAssertNil(unboxedFromInvalidDictionary, "Unbox did not return nil for an invalid dictionary")
    }
    
    func testWithInvalidOptionalValue() {
        var validDictionary = UnboxTestDictionaryWithAllRequiredKeysWithValidValues(false)
        validDictionary[UnboxTestMock.optionalBoolKey] = "Not a Bool"
        
        let unboxed: UnboxTestMock? = Unbox(validDictionary)
        XCTAssertNotNil(unboxed, "Invalid optional values should be ignored")
    }
    
    func testUnboxingFromValidData() {
        let dictionary = UnboxTestDictionaryWithAllRequiredKeysWithValidValues(false)
        
        do {
            let data = try NSJSONSerialization.dataWithJSONObject(dictionary, options: [])
            let unboxed: UnboxTestMock? = Unbox(data)
            XCTAssertNotNil(unboxed, "Could not unbox from data")
        } catch {
            XCTFail("Could not decode data from dictionary: \(dictionary)")
        }
    }
    
    func testUnboxingArrayOfDictionaries() {
        let dictionaries = [
            UnboxTestDictionaryWithAllRequiredKeysWithValidValues(false),
            UnboxTestDictionaryWithAllRequiredKeysWithValidValues(false),
            UnboxTestDictionaryWithAllRequiredKeysWithValidValues(false)
        ]
        
        guard let unboxedArray: [UnboxTestMock] = Unbox(dictionaries) else {
            return XCTFail()
        }
        
        XCTAssertEqual(unboxedArray.count, 3)
        
        for unboxed in unboxedArray {
            unboxed.verifyAgainstDictionary(UnboxTestDictionaryWithAllRequiredKeysWithValidValues(false))
        }
    }
    
    func testThrowingForMissingRequiredValues() {
        let validDictionary = UnboxTestDictionaryWithAllRequiredKeysWithValidValues(false)
        
        for key in validDictionary.keys {
            var invalidDictionary = validDictionary
            invalidDictionary.removeValueForKey(key)
            
            do {
                try UnboxOrThrow(invalidDictionary) as UnboxTestMock
                XCTFail("Unbox should have thrown for a missing value")
            } catch UnboxError.MissingKey(key) {
                // Test passed
            } catch {
                XCTFail("Unbox did not return the correct error type")
            }
        }
    }
    
    func testThrowingForInvalidRequiredValues() {
        let validDictionary = UnboxTestDictionaryWithAllRequiredKeysWithValidValues(false)
        
        for key in validDictionary.keys {
            var invalidDictionary = validDictionary
            let invalidValue = NSObject()
            let invalidValueDescription = "\(invalidValue)"
            invalidDictionary[key] = invalidValue
            
            do {
                try UnboxOrThrow(invalidDictionary) as UnboxTestMock
                XCTFail("Unbox should have thrown for an invalid value")
            } catch UnboxError.InvalidValue(key, invalidValueDescription) {
                // Test passed
            } catch {
                XCTFail("Unbox did not return the correct error type")
            }
            
            let unboxed: UnboxTestMock? = Unbox(invalidDictionary)
            XCTAssertNil(unboxed, "Unbox did not return nil for an invalid dictionary")
        }
    }
    
    func testThrowingForInvalidData() {
        if let data = "Not a dictionary".dataUsingEncoding(NSUTF8StringEncoding) {
            do {
                try UnboxOrThrow(data) as UnboxTestMock
                XCTFail("Unbox should have thrown for invalid data")
            } catch UnboxError.InvalidData {
                // Test passed
            } catch {
                XCTFail("Unbox did not return the correct error type")
            }
        } else {
            XCTFail("Could not create data from a string")
        }
    }
    
    func testThrowingForInvalidDataArray() {
        let notDictionaryArray = [12, 13, 9]
        
        guard let data = try? NSJSONSerialization.dataWithJSONObject(notDictionaryArray, options: []) else {
            return XCTFail()
        }
        
        do {
            try UnboxOrThrow(data) as UnboxTestMock
            XCTFail()
        } catch UnboxError.InvalidData {
            // Test passed
        } catch {
            XCTFail("Unbox did not return the correct error type")
        }
    }
    
    func testThrowingForSingleInvalidDictionaryInArray() {
        let dictionaries = [
            UnboxTestDictionaryWithAllRequiredKeysWithValidValues(false),
            ["invalid" : "dictionary"],
            UnboxTestDictionaryWithAllRequiredKeysWithValidValues(false)
        ]
        
        do {
            try UnboxOrThrow(dictionaries) as [UnboxTestMock]
            XCTFail()
        } catch {
            // Test passed
        }
    }
    
    func testContext() {
        class Model: Unboxable {
            let nestedUnboxable: Model?
            
            required init(unboxer: Unboxer) {
                if let context = unboxer.context as? String {
                    XCTAssertTrue("context" == context, "")
                } else {
                    XCTFail("Context was of an unexpected type: \(unboxer.context)")
                }
                
                self.nestedUnboxable = unboxer.unbox("nested")
            }
        }
        
        let unboxed: Model? = Unbox(["nested" : UnboxableDictionary()], context: "context")
        
        XCTAssertFalse(unboxed == nil, "Could not unbox with a context")
    }
    
    func testRequiredContext() {
        let dictionary = [
            "nested" : [:],
            "nestedArray": [[:]]
        ]
        
        if let model: UnboxTestContextMock = Unbox(dictionary, context: "context") {
            XCTAssertEqual(model.context, "context")
            
            if let nestedModel = model.nested {
                XCTAssertEqual(nestedModel.context, "nestedContext")
            } else {
                XCTFail("Failed to unbox nested model")
            }
            
            if let nestedArrayModel = model.nestedArray?.first {
                XCTAssertEqual(nestedArrayModel.context, "nestedArrayContext")
            } else {
                XCTFail("Failed to unbox nested model array")
            }
        } else {
            XCTFail("Failed to unbox")
        }
    }

    func testAccessingNestedDictionaryWithKeyPath() {
        struct KeyPathModel: Unboxable {
            let intValue: Int
            let dictionary: UnboxableDictionary

            init(unboxer: Unboxer) {
                let intKeyPathComponents = [UnboxTestMock.requiredUnboxableDictionaryKey, "test", UnboxTestMock.requiredIntKey]
                let keyPath = intKeyPathComponents.joinWithSeparator(".")
                self.intValue = unboxer.unbox(keyPath)

                let dictionaryKeyPath = [UnboxTestMock.requiredUnboxableDictionaryKey, "test"].joinWithSeparator(".")
                self.dictionary = unboxer.unbox(dictionaryKeyPath)
            }
        }


        let validDictionary = UnboxTestDictionaryWithAllRequiredKeysWithValidValues(false)
        let model: KeyPathModel? = Unbox(validDictionary)
        XCTAssertNotNil(model)
        XCTAssertEqual(15, model?.intValue)
        if let result = model?.dictionary[UnboxTestMock.requiredArrayKey] as? [String] {
            XCTAssertEqual(["unbox", "is", "pretty", "cool", "right?"], result)
        } else {
            XCTFail()
        }
    }
    
    func testAllKeys() {
        let dictionary = [
            "a" : "A",
            "b" : 12,
            "c" : ["C"]
        ]
        
        struct Model: Unboxable {
            init(unboxer: Unboxer) {
                XCTAssertEqual(unboxer.allKeys.count, 3)
                XCTAssertTrue(unboxer.allKeys.contains("a"))
                XCTAssertTrue(unboxer.allKeys.contains("b"))
                XCTAssertTrue(unboxer.allKeys.contains("c"))
            }
        }
        
        Unbox(dictionary) as Model?
    }
}

private func UnboxTestDictionaryWithAllRequiredKeysWithValidValues(nested: Bool) -> UnboxableDictionary {
    var dictionary: UnboxableDictionary = [
        UnboxTestMock.requiredBoolKey : true,
        UnboxTestMock.requiredIntKey : 15,
        UnboxTestMock.requiredDoubleKey : Double(1.5),
        UnboxTestMock.requiredFloatKey : Float(3.14),
        UnboxTestMock.requiredCGFloatKey : 0.72,
        UnboxTestMock.requiredEnumKey : 1,
        UnboxTestMock.requiredStringKey :  "hello",
        UnboxTestMock.requiredURLKey : "http://www.google.com",
        UnboxTestMock.requiredArrayKey : ["unbox", "is", "pretty", "cool", "right?"]
    ]
    
    if !nested {
        dictionary[UnboxTestMock.requiredUnboxableKey] = UnboxTestDictionaryWithAllRequiredKeysWithValidValues(true)
        dictionary[UnboxTestMock.requiredUnboxableArrayKey] = [UnboxTestDictionaryWithAllRequiredKeysWithValidValues(true)]
        dictionary[UnboxTestMock.requiredUnboxableDictionaryKey] = ["test" : UnboxTestDictionaryWithAllRequiredKeysWithValidValues(true)]
    }
    
    return dictionary
}

// MARK: - Mocks

private enum UnboxTestEnum: Int, UnboxableEnum {
    case First
    case Second
    
    private static func unboxFallbackValue() -> UnboxTestEnum {
        return .First
    }
}

private class UnboxTestBaseMock: Unboxable {
    static let requiredBoolKey = "requiredBool"
    static let optionalBoolKey = "optionalBool"
    static let requiredIntKey = "requiredInt"
    static let optionalIntKey = "optionalInt"
    static let requiredDoubleKey = "requiredDouble"
    static let optionalDoubleKey = "optionalDouble"
    static let requiredFloatKey = "requiredFloat"
    static let optionalFloatKey = "optionalFloat"
    static let requiredCGFloatKey = "requiredCGFloat"
    static let optionalCGFloatKey = "optionalCGFloat"
    static let requiredEnumKey = "requiredEnum"
    static let optionalEnumKey = "optionalEnum"
    static let requiredStringKey = "requiredString"
    static let optionalStringKey = "optionalString"
    static let requiredURLKey = "requiredURL"
    static let optionalURLKey = "optionalURL"
    static let requiredArrayKey = "requiredArray"
    static let optionalArrayKey = "optionalArray"
    
    let requiredBool: Bool
    let optionalBool: Bool?
    let requiredInt: Int
    let optionalInt: Int?
    let requiredDouble: Double
    let optionalDouble: Double?
    let requiredFloat: Float
    let optionalFloat: Float?
    let requiredCGFloat: CGFloat
    let optionalCGFloat: CGFloat?
    let requiredEnum: UnboxTestEnum
    let optionalEnum: UnboxTestEnum?
    let requiredString: String
    let optionalString: String?
    let requiredURL: NSURL
    let optionalURL: NSURL?
    let requiredArray: [String]
    let optionalArray: [String]?
    
    required init(unboxer: Unboxer) {
        self.requiredBool = unboxer.unbox(UnboxTestBaseMock.requiredBoolKey)
        self.optionalBool = unboxer.unbox(UnboxTestBaseMock.optionalBoolKey)
        self.requiredInt = unboxer.unbox(UnboxTestBaseMock.requiredIntKey)
        self.optionalInt = unboxer.unbox(UnboxTestBaseMock.optionalIntKey)
        self.requiredDouble = unboxer.unbox(UnboxTestBaseMock.requiredDoubleKey)
        self.optionalDouble = unboxer.unbox(UnboxTestBaseMock.optionalDoubleKey)
        self.requiredFloat = unboxer.unbox(UnboxTestBaseMock.requiredFloatKey)
        self.optionalFloat = unboxer.unbox(UnboxTestBaseMock.optionalFloatKey)
        self.requiredCGFloat = unboxer.unbox(UnboxTestBaseMock.requiredCGFloatKey)
        self.optionalCGFloat = unboxer.unbox(UnboxTestBaseMock.optionalCGFloatKey)
        self.requiredEnum = unboxer.unbox(UnboxTestBaseMock.requiredEnumKey)
        self.optionalEnum = unboxer.unbox(UnboxTestBaseMock.optionalEnumKey)
        self.requiredString = unboxer.unbox(UnboxTestBaseMock.requiredStringKey)
        self.optionalString = unboxer.unbox(UnboxTestBaseMock.optionalStringKey)
        self.requiredURL = unboxer.unbox(UnboxTestBaseMock.requiredURLKey)
        self.optionalURL = unboxer.unbox(UnboxTestBaseMock.optionalURLKey)
        self.requiredArray = unboxer.unbox(UnboxTestBaseMock.requiredArrayKey)
        self.optionalArray = unboxer.unbox(UnboxTestBaseMock.optionalArrayKey)
    }
    
    func verifyAgainstDictionary(dictionary: UnboxableDictionary) {
        for (key, value) in dictionary {
            let verificationOutcome: Bool
            
            switch key {
            case UnboxTestBaseMock.requiredBoolKey:
                verificationOutcome = self.verifyPropertyValue(self.requiredBool, againstDictionaryValue: value)
            case UnboxTestBaseMock.optionalBoolKey:
                verificationOutcome = self.verifyPropertyValue(self.optionalBool, againstDictionaryValue: value)
            case UnboxTestBaseMock.requiredIntKey:
                verificationOutcome = self.verifyPropertyValue(self.requiredInt, againstDictionaryValue: value)
            case UnboxTestBaseMock.optionalIntKey:
                verificationOutcome = self.verifyPropertyValue(self.optionalInt, againstDictionaryValue: value)
            case UnboxTestBaseMock.requiredDoubleKey:
                verificationOutcome = self.verifyPropertyValue(self.requiredDouble, againstDictionaryValue: value)
            case UnboxTestBaseMock.optionalDoubleKey:
                verificationOutcome = self.verifyPropertyValue(self.optionalDouble, againstDictionaryValue: value)
            case UnboxTestBaseMock.requiredFloatKey:
                verificationOutcome = self.verifyPropertyValue(self.requiredFloat, againstDictionaryValue: value)
            case UnboxTestBaseMock.optionalFloatKey:
                verificationOutcome = self.verifyPropertyValue(self.optionalFloat, againstDictionaryValue: value)
            case UnboxTestBaseMock.requiredCGFloatKey:
                verificationOutcome = self.verifyPropertyValue(self.requiredCGFloat, againstDictionaryValue: value)
            case UnboxTestBaseMock.optionalCGFloatKey:
                verificationOutcome = self.verifyPropertyValue(self.optionalCGFloat, againstDictionaryValue: value)
            case UnboxTestBaseMock.requiredEnumKey:
                verificationOutcome = self.verifyEnumPropertyValue(self.requiredEnum, againstDictionaryValue: value)
            case UnboxTestBaseMock.optionalEnumKey:
                verificationOutcome = self.verifyEnumPropertyValue(self.optionalEnum, againstDictionaryValue: value)
            case UnboxTestBaseMock.requiredStringKey:
                verificationOutcome = self.verifyPropertyValue(self.requiredString, againstDictionaryValue: value)
            case UnboxTestBaseMock.optionalStringKey:
                verificationOutcome = self.verifyPropertyValue(self.optionalString, againstDictionaryValue: value)
            case UnboxTestBaseMock.requiredURLKey:
                verificationOutcome = self.verifyURLPropertyValue(self.requiredURL, againstDictionaryValue: value)
            case UnboxTestBaseMock.optionalURLKey:
                verificationOutcome = self.verifyURLPropertyValue(self.optionalURL, againstDictionaryValue: value)
            case UnboxTestBaseMock.requiredArrayKey:
                verificationOutcome = self.verifyArrayPropertyValue(self.requiredArray, againstDictionaryValue: value)
            case UnboxTestBaseMock.optionalArrayKey:
                verificationOutcome = self.verifyArrayPropertyValue(self.optionalArray, againstDictionaryValue: value)
            default:
                verificationOutcome = true
            }
            
            XCTAssertTrue(verificationOutcome, "Verification failed for key: " + key)
        }
    }
    
    func verifyPropertyValue<T: Equatable>(propertyValue: T?, againstDictionaryValue dictionaryValue: AnyObject?) -> Bool {
        if let propertyValue = propertyValue {
            if let typedDictionaryValue = dictionaryValue as? T {
                return propertyValue == typedDictionaryValue
            }
        }
        
        return false
    }
    
    func verifyEnumPropertyValue<T: UnboxableEnum where T: Equatable>(propertyValue: T?, againstDictionaryValue dictionaryValue: AnyObject?) -> Bool {
        if let rawValue = dictionaryValue as? T.RawValue {
            if let enumValue = T(rawValue: rawValue) {
                return propertyValue == enumValue
            }
        }
        
        return false
    }
    
    func verifyURLPropertyValue(propertyValue: NSURL?, againstDictionaryValue dictionaryValue: AnyObject?) -> Bool {
        if let string = dictionaryValue as? String {
            return self.verifyPropertyValue(self.requiredURL, againstDictionaryValue: NSURL(string: string))
        }
        
        return false
    }
    
    func verifyArrayPropertyValue<T: Equatable>(propertyValue: [T]?, againstDictionaryValue dictionaryValue: AnyObject?) -> Bool {
        if let propertyValue = propertyValue {
            if let dictionaryArrayValue = dictionaryValue as? [T] {
                for i in 0..<dictionaryArrayValue.count {
                    if dictionaryArrayValue[i] != propertyValue[i] {
                        return false
                    }
                }
                
                return true
            }
        }
        
        return false
    }
}

private class UnboxTestMock: UnboxTestBaseMock {
    static let requiredUnboxableKey = "requiredUnboxable"
    static let optionalUnboxableKey = "optionalUnboxable"
    static let requiredUnboxableArrayKey = "requiredUnboxableArray"
    static let optionalUnboxableArrayKey = "optionalUnboxableArray"
    static let requiredUnboxableDictionaryKey = "requiredUnboxableDictionary"
    static let optionalUnboxableDictionaryKey = "optionalUnboxableDictionary"
    
    let requiredUnboxable: UnboxTestBaseMock
    let optionalUnboxable: UnboxTestBaseMock?
    let requiredUnboxableArray: [UnboxTestBaseMock]
    let optionalUnboxableArray: [UnboxTestBaseMock]?
    let requiredUnboxableDictionary: [String : UnboxTestBaseMock]
    let optionalUnboxableDictionary: [String : UnboxTestBaseMock]?
    
    required init(unboxer: Unboxer) {
        self.requiredUnboxable = unboxer.unbox(UnboxTestMock.requiredUnboxableKey)
        self.optionalUnboxable = unboxer.unbox(UnboxTestMock.optionalUnboxableKey)
        self.requiredUnboxableArray = unboxer.unbox(UnboxTestMock.requiredUnboxableArrayKey)
        self.optionalUnboxableArray = unboxer.unbox(UnboxTestMock.optionalUnboxableArrayKey)
        self.requiredUnboxableDictionary = unboxer.unbox(UnboxTestMock.requiredUnboxableDictionaryKey)
        self.optionalUnboxableDictionary = unboxer.unbox(UnboxTestMock.optionalUnboxableDictionaryKey)
        
        super.init(unboxer: unboxer)
    }
}

private final class UnboxTestContextMock: UnboxableWithContext {
    let context: String
    let nested: UnboxTestContextMock?
    let nestedArray: [UnboxTestContextMock]?
    
    init(unboxer: Unboxer, context: String) {
        self.context = context
        self.nested = unboxer.unbox("nested", context: "nestedContext")
        self.nestedArray = unboxer.unbox("nestedArray", context: "nestedArrayContext")
    }
}
